# -*- coding: utf-8 -*-

"""
Prompter post-processing module

This module contains functions that help filter LLMs results to get the final anomalies.
"""

import numpy as np


def val2idx(y, X):
    """Convert detected anomalies values into indices.

    Convert windows of detected anomalies values into an array of all indices
    in the input sequence that have those values.

    Args:
        y (ndarray):
            A 3d array containing detected anomalous values from different
            responses of each window.
        X (ndarray):
            rolling window sequences.
    Returns:
        List([ndarray]):
            A 3d array containing detected anomalous indices from different
            responses of each window.
    """

    idx_list = []
    for anomalies_list, seq in zip(y, X):
        idx_win_list = []
        for anomalies in anomalies_list:
            mask = np.isin(seq, anomalies)
            indices = np.where(mask)[0]
            idx_win_list.append(indices)
        idx_list.append(idx_win_list)
    idx_list = np.array(idx_list, dtype=object)
    return idx_list


def find_anomalies_in_windows(y, alpha=0.5):
    """Get the final list of anomalous indices of each window

    Choose anomalous index in the sequence based on multiple LLM responses

    Args:
        y (ndarray):
            A 3d array containing detected anomalous values from different
            responses of each window.
        alpha (float):
            Percent of votes needed for an index to be anomalous. Default to `0.5`.

    Returns:
        ndarray:
            A 2-dimensional array containing final anomalous indices of each windows.
    """

    idx_list = []
    for samples in y:
        min_vote = np.ceil(alpha * len(samples))
        # print(type(samples.tolist()))

        flattened_res = np.concatenate(samples.tolist())

        unique_elements, counts = np.unique(flattened_res, return_counts=True)

        final_list = unique_elements[counts >= min_vote]

        idx_list.append(final_list)
    idx_list = np.array(idx_list, dtype=object)
    return idx_list


def merge_anomalous_sequences(y, first_index, window_size, step_size, beta=0.5):
    """Get the final list of anomalous indices of a sequence when merging all rolling windows

    Args:
        y (ndarray):
            A 2-dimensional array containing anomalous indices of each window.
        first_index (ndarray):
            A 1-dimensional array contaning the first index of each window.
        window_size (int):
            Length of each window
        step_size (int):
            Indicating the number of steps the window moves forward each round.
        beta (float):
            Percent of windows needed for index to be anomalous. Default to `0.5`.

    Return:
        ndarray:
            A 1-dimensional array containing final anomalous indices.
    """
    anomalies = [arr + first_idx for (arr, first_idx) in zip(y, first_index)]

    min_vote = np.ceil(beta * window_size / step_size)

    flattened_res = np.concatenate(anomalies)

    unique_elements, counts = np.unique(flattened_res, return_counts=True)

    final_list = unique_elements[counts >= min_vote]

    return np.sort(final_list)


def format_anomalies(y, timestamp, padding_size=50):
    """Convert list of anomalous indices to list of intervals by padding to both sides
    and merge overlapping

    Args:
        y (ndarray):
            A 1-dimensional array of indices.
        timestamp (ndarray):
            List of full timestamp of the signal
        padding_size (int):
            Number of steps to pad on both sides of a timestamp point. Default to `50`.

    Returns:
        List[Tuple]:
            List of intervals (start, end, score).
    """
    y = timestamp[y]  # Convert list of indices into list of timestamps
    start, end = timestamp[0], timestamp[-1]
    interval = timestamp[1] - timestamp[0]
    intervals = []
    for timestamp in y:
        intervals.append((max(start, timestamp - padding_size * interval),
                         min(end, timestamp + padding_size * interval)))
    if not intervals:
        return []

    intervals.sort(key=lambda x: x[0])  # Sort intervals based on start time
    merged_intervals = [intervals[0]]  # Initialize merged intervals with the first interval

    for current_interval in intervals[1:]:
        previous_interval = merged_intervals[-1]

        # If the current interval overlaps with the previous one, merge them
        if current_interval[0] <= previous_interval[1]:
            previous_interval = (
                previous_interval[0], max(
                    previous_interval[1], current_interval[1]))
            merged_intervals[-1] = previous_interval
        else:
            merged_intervals.append(current_interval)  # Append the current interval if no overlap

    merged_intervals = [(interval[0], interval[1], 0) for interval in merged_intervals]
    return merged_intervals



# -*- coding: utf-8 -*-

import numpy as np
from pytest import fixture

from sigllm.primitives.prompting.anomalies import (
    find_anomalies_in_windows, format_anomalies, merge_anomalous_sequences, val2idx,)


@fixture
def anomaly_list_within_seq():
    return np.array([[np.array([0, 3]), np.array([1]), np.array([1, 2])],
                     [np.array([0]), np.array([1, 4]), np.array([2, 3])],
                     [np.array([0, 2]), np.array([]), np.array([0, 1])]], dtype=object)


@fixture
def anomaly_list_across_seq():
    return np.array([np.array([0]),
                     np.array([1, 2]),
                     np.array([0, 2]),
                     np.array([1, 2]),
                     np.array([1])], dtype=object)


@fixture
def first_indices():
    return np.array([0, 1, 2, 3, 4])


@fixture
def window_size():
    return 3


@fixture
def step_size():
    return 1


@fixture
def idx_list():
    return np.array([32, 545, 689, 1103, 1134])


@fixture
def anomalous_val():
    return np.array([[np.array([0, 3]), np.array([])],
                    [np.array([2]), np.array([4])]], dtype=object)


@fixture
def windows():
    return np.array([[0, 1, 0, 3],
                     [3, 2, 6, 2]])


@fixture
def timestamp():
    return np.array(range(1000, 13000, 10))


def test_ano_within_windows(anomaly_list_within_seq):
    expected = np.array([np.array([1]),
                         np.array([]),
                         np.array([0])], dtype=object)

    result = find_anomalies_in_windows(anomaly_list_within_seq)

    for r, e in zip(result, expected):
        np.testing.assert_equal(r, e)


def test_merge_anomaly_seq(anomaly_list_across_seq, first_indices, window_size, step_size):
    expected = np.array([2, 4, 5])

    result = merge_anomalous_sequences(
        anomaly_list_across_seq,
        first_indices,
        window_size,
        step_size)

    np.testing.assert_equal(result, expected)


# val2idx
def test_val2idx(anomalous_val, windows):
    expected = np.array([[np.array([0, 2, 3]), np.array([])],
                         [np.array([1, 3]), np.array([])]], dtype=object)
    result = val2idx(anomalous_val, windows)

    for r_list, e_list in zip(result, expected):
        for r, e in zip(r_list, e_list):
            np.testing.assert_equal(r, e)

# timestamp2interval


def test_format_anomalies(idx_list, timestamp):
    expected = [(1000, 1820, 0), (5950, 6950, 0), (7390, 8390, 0),
                (11530, 12840, 0)]
    result = format_anomalies(idx_list, timestamp)

    assert expected == result



# -*- coding: utf-8 -*-

import json
import logging
import os

import torch
from tqdm import tqdm
from transformers import AutoModelForCausalLM, AutoTokenizer

PROMPT_PATH = os.path.join(
    os.path.dirname(os.path.abspath(__file__)),
    'huggingface_messages.json'
)

PROMPTS = json.load(open(PROMPT_PATH))

LOGGER = logging.getLogger(__name__)

DEFAULT_BOS_TOKEN = "<s>"
DEFAULT_EOS_TOKEN = "</s>"
DEFAULT_UNK_TOKEN = "<unk>"
DEFAULT_PAD_TOKEN = "<pad>"

VALID_NUMBERS = list("0123456789")

DEFAULT_MODEL = 'mistralai/Mistral-7B-Instruct-v0.2'


class HF:
    """Prompt Pretrained models on HuggingFace to detect anomalies in a time series.

    Args:
        name (str):
            Model name. Default to `'mistralai/Mistral-7B-Instruct-v0.2'`.
        sep (str):
            String to separate each element in values. Default to `','`.
        anomalous_percent (float):
            Expected percentage of time series that are anomalous. Default to `0.5`.
        temp (float):
            The value used to modulate the next token probabilities. Default to `1`.
        top_p (float):
            If set to float < 1, only the smallest set of most probable tokens with
            probabilities that add up to `top_p` or higher are kept for generation.
            Default to `1`.
        raw (bool):
            Whether to return the raw output or not. Defaults to `False`.
        samples (int):
            Number of responsed to generate for each input message. Default to `10`.
        padding (int):
            Additional padding token to forecast to reduce short horizon predictions.
            Default to `0`.
    """

    def __init__(self, name=DEFAULT_MODEL, sep=',', anomalous_percent=0.5, temp=1, top_p=1,
                 raw=False, samples=10, padding=0):
        self.name = name
        self.sep = sep
        self.anomalous_percent = anomalous_percent
        self.temp = temp
        self.top_p = top_p
        self.raw = raw
        self.samples = samples
        self.padding = padding

        self.tokenizer = AutoTokenizer.from_pretrained(self.name, use_fast=False)

        # special tokens
        special_tokens_dict = dict()
        if self.tokenizer.eos_token is None:
            special_tokens_dict["eos_token"] = DEFAULT_EOS_TOKEN
        if self.tokenizer.bos_token is None:
            special_tokens_dict["bos_token"] = DEFAULT_BOS_TOKEN
        if self.tokenizer.unk_token is None:
            special_tokens_dict["unk_token"] = DEFAULT_UNK_TOKEN
        if self.tokenizer.pad_token is None:
            special_tokens_dict["pad_token"] = DEFAULT_PAD_TOKEN

        self.tokenizer.add_special_tokens(special_tokens_dict)
        self.tokenizer.pad_token = self.tokenizer.eos_token  # indicate the end of the time series

        # invalid tokens
        valid_tokens = []
        for number in VALID_NUMBERS:
            token = self.tokenizer.convert_tokens_to_ids(number)
            valid_tokens.append(token)

        valid_tokens.append(self.tokenizer.convert_tokens_to_ids(self.sep))
        self.invalid_tokens = [[i]
                               for i in range(len(self.tokenizer) - 1) if i not in valid_tokens]

        self.model = AutoModelForCausalLM.from_pretrained(
            self.name,
            device_map="auto",
            torch_dtype=torch.float16,
        )

        self.model.eval()

    def detect(self, X, **kwargs):
        """Use HF to detect anomalies of a signal.

        Args:
            X (ndarray):
                Input sequences of strings containing signal values

        Returns:
            list, list:
                * List of detected anomalous values.
                * Optionally, a list of dictionaries for raw output.
        """

        input_length = len(self.tokenizer.encode(X[0].flatten().tolist()[0]))
        max_tokens = input_length * float(self.anomalous_percent)
        all_responses, all_generate_ids = [], []

        for text in tqdm(X):
            system_message = PROMPTS['system_message']
            user_message = PROMPTS['user_message']
            message = ' '.join([system_message, user_message, text, '[RESPONSE]'])

            input_length = len(self.tokenizer.encode(message[0]))

            tokenized_input = self.tokenizer(
                message,
                return_tensors="pt"
            ).to("cuda")

            generate_ids = self.model.generate(
                **tokenized_input,
                do_sample=True,
                max_new_tokens=max_tokens,
                temperature=self.temp,
                top_p=self.top_p,
                bad_words_ids=self.invalid_tokens,
                renormalize_logits=True,
                num_return_sequences=self.samples
            )

            responses = self.tokenizer.batch_decode(
                generate_ids[:, input_length:],
                skip_special_tokens=True,
                clean_up_tokenization_spaces=False
            )
            all_responses.append(responses)
            all_generate_ids.append(generate_ids)

        if self.raw:
            return all_responses, all_generate_ids

        return all_responses
